from configuration import get_object_configuration_mapper
from flask import Request
from inuits_policy_based_auth.authentication.base_authentication_policy import (
    BaseAuthenticationPolicy,
)
from os import getenv
from storage.storagemanager import StorageManager


class XUserHeadersPolicy(BaseAuthenticationPolicy):
    def authenticate(self, user_context, request_context):
        request: Request = request_context.http_request

        auth_header = request.headers.get("Authorization", "")
        if auth_header.startswith("Bearer "):
            token = auth_header.removeprefix("Bearer ")
            static_jwt = getenv("STATIC_JWT")

            if static_jwt and token == static_jwt:
                user_email = request.headers.get("X-User-Email")
                if user_email:
                    config = get_object_configuration_mapper().get("user")
                    storage_manager = StorageManager()
                    user = (
                        storage_manager.get_db_engine().get_item_from_collection_by_id(
                            config.crud()["collection"], user_email
                        )
                    )

                    if user:
                        user_context.id = user.get("_id", user.get("id"))
                        user_context.email = user_email

        return user_context



from datetime import datetime, timezone
from enum import Enum
from os import getenv
from typing import Literal

from elody.object_configurations.elody_configuration import (
    ElodyConfiguration,
)
from elody.util import send_cloudevent


class Status(str, Enum):
    QUEUED = "queued"
    RUNNING = "running"
    FINISHED = "finished"
    FAILED = "failed"
    WARNING = "warning"


class JobConfiguration(ElodyConfiguration):
    SCHEMA_TYPE = "elody"
    SCHEMA_VERSION = 1

    def crud(self):
        crud = {
            "collection": "jobs",
            "collection_history": "",
            "add_document_to_job": lambda *args, **kwargs: self._add_document_to_job(
                *args, **kwargs
            ),
            "init_job": lambda *args, **kwargs: self._init_job(*args, **kwargs),
            "start_job": lambda *args, **kwargs: self._start_job(*args, **kwargs),
            "finish_job": lambda *args, **kwargs: self._finish_job(*args, **kwargs),
            "finish_job_with_warning": lambda *args, **kwargs: self._finish_job_with_warning(
                *args, **kwargs
            ),
            "fail_job": lambda *args, **kwargs: self._fail_job(*args, **kwargs),
            "handle_parent_job_finished": lambda *args, **kwargs: self._handle_parent_job_finished(
                *args, **kwargs
            ),
        }
        return {**super().crud(), **crud}

    def document_info(self):
        return super().document_info()

    def logging(self, flat_document, **kwargs):
        return super().logging(flat_document, **kwargs)

    def migration(self):
        return super().migration()

    def serialization(self, from_format, to_format):
        return super().serialization(from_format, to_format)

    def validation(self):
        return super().validation()

    def _creator(self, post_body, *, get_user_context={}, **_):
        document = super()._creator(post_body)
        if email := get_user_context().email:
            document["last_editor"] = email
        return document

    def _post_crud_hook(self, *, crud, document, get_rabbit, **kwargs):
        if crud == "create":
            send_cloudevent(
                get_rabbit(),
                getenv("MQ_EXCHANGE", "dams"),
                "dams.job_created",
                document,
            )
        elif crud == "update":
            send_cloudevent(
                get_rabbit(),
                getenv("MQ_EXCHANGE", "dams"),
                "dams.job_changed",
                document,
            )

    def _add_document_to_job(
        self,
        id,
        id_of_document_job_was_initiated_for,
        *,
        get_rabbit,
    ):
        relations = []
        if id_of_document_job_was_initiated_for:
            relations.append(
                {"key": id_of_document_job_was_initiated_for, "type": "isJobFor"}
            )
        document = {
            "id": id,
            "patch": {
                "relations": (relations),
            },
        }
        self._post_crud_hook(crud="update", document=document, get_rabbit=get_rabbit)

    def _init_job(
        self,
        name,
        job_type,
        *,
        get_rabbit,
        get_user_context=None,
        user_email=None,
        parent_id=None,
        id_of_document_job_was_initiated_for=None,
        track_async_children: bool | None = None,
    ) -> str:
        relations = []
        if parent_id:
            relations.append({"key": parent_id, "type": "hasParentJob"})
        if id_of_document_job_was_initiated_for:
            relations.append(
                {"key": id_of_document_job_was_initiated_for, "type": "isJobFor"}
            )

        metadata = [
            {"key": "name", "value": name},
            {"key": "status", "value": Status.QUEUED.value},
            {"key": "type", "value": job_type},
        ]

        if track_async_children:
            metadata.append(
                {
                    "key": "child_jobs",
                    "value": {
                        Status.QUEUED.value: 0,
                        Status.RUNNING.value: 0,
                        Status.FAILED.value: 0,
                        Status.FINISHED.value: 0,
                        Status.WARNING.value: 0,
                    },
                }
            )

        job = self.crud()["creator"](
            {
                "metadata": metadata,
                "relations": relations,
                "type": "job",
            },
            get_user_context=get_user_context
            or (lambda: type("UserContext", (object,), {"email": user_email})()),
        )

        self._post_crud_hook(
            crud="create", document=job, parent_id=parent_id, get_rabbit=get_rabbit
        )
        return job["_id"]

    def _start_job(
        self,
        id,
        *,
        get_rabbit,
    ):
        document = {
            "id": id,
            "patch": {
                "started_at": datetime.now(timezone.utc),
                "metadata": [{"key": "status", "value": Status.RUNNING.value}],
            },
        }
        self._post_crud_hook(crud="update", document=document, get_rabbit=get_rabbit)

    def _finish_job(
        self,
        id,
        *,
        get_rabbit,
    ):
        document = {
            "id": id,
            "patch": {
                "metadata": [{"key": "status", "value": Status.FINISHED.value}],
            },
        }
        self._post_crud_hook(crud="update", document=document, get_rabbit=get_rabbit)

    def _finish_job_with_warning(
        self,
        id,
        *,
        get_rabbit,
        info_message=None,
    ):
        document = {
            "id": id,
            "patch": {
                "metadata": [
                    {"key": "status", "value": Status.FINISHED.value},
                    {"key": "info", "value": info_message},
                ],
            },
        }
        self._post_crud_hook(crud="update", document=document, get_rabbit=get_rabbit)

    def _fail_job(self, id, exception_message, *, get_rabbit):

        status, message = self.__handle_error_warnings(exception_message)

        document = {
            "id": id,
            "patch": {
                "metadata": [
                    {"key": "info", "value": message},
                    {
                        "key": "status",
                        "value": (status.value),
                    },
                ]
            },
        }
        self._post_crud_hook(crud="update", document=document, get_rabbit=get_rabbit)

    def _handle_parent_job_finished(self, id, parent_child_status, *, get_rabbit):

        child_jobs_failed = parent_child_status[Status.FAILED.value]
        child_jobs_warning = parent_child_status[Status.WARNING.value]

        if child_jobs_failed == 0:
            if child_jobs_warning:
                self.crud()["finish_job_with_warning"](
                    id, get_rabbit=get_rabbit, info_message="Some subjobs have warnings"
                )
            self.crud()["finish_job"](id, get_rabbit=get_rabbit)
        else:
            self.crud()["fail_job"](
                id, exception_message="some subjobs failed", get_rabbit=get_rabbit
            )

    def __handle_error_warnings(
        self,
        errormessage: str,
    ) -> tuple[Literal[Status.WARNING] | Literal[Status.FAILED], str]:

        if errormessage.startswith("W4009"):
            return Status.WARNING, errormessage

        return Status.FAILED, errormessage



try:
    from config import get_object_configuration_mapper

    _config = get_object_configuration_mapper().get("job")
except ModuleNotFoundError:
    from elody.object_configurations.job_configuration import JobConfiguration

    _config = JobConfiguration()


def add_document_to_job(
    id,
    id_of_document_job_was_initiated_for,
    *,
    get_rabbit,
):
    _config.crud()["add_document_to_job"](
        id=id,
        id_of_document_job_was_initiated_for=id_of_document_job_was_initiated_for,
        get_rabbit=get_rabbit,
    )


def init_job(
    name,
    job_type,
    *,
    get_rabbit,
    get_user_context=None,
    user_email=None,
    parent_id=None,
    id_of_document_job_was_initiated_for=None,
    track_async_children: bool | None = None,
) -> str:
    return _config.crud()["init_job"](
        name=name,
        job_type=job_type,
        get_rabbit=get_rabbit,
        get_user_context=get_user_context,
        user_email=user_email,
        parent_id=parent_id,
        id_of_document_job_was_initiated_for=id_of_document_job_was_initiated_for,
        track_async_children=track_async_children,
    )


def start_job(
    id,
    *,
    get_rabbit,
):
    _config.crud()["start_job"](
        id=id,
        get_rabbit=get_rabbit,
    )


def finish_job(
    id,
    *,
    get_rabbit,
):
    _config.crud()["finish_job"](
        id=id,
        get_rabbit=get_rabbit,
    )


def soft_finish_job(
    id,
    *,
    get_rabbit,
):
    _config.crud()["soft_finish_job"](id=id, get_rabbit=get_rabbit)


def fail_job(id, exception_message, *, get_rabbit):
    _config.crud()["fail_job"](
        id=id, exception_message=exception_message, get_rabbit=get_rabbit
    )


def handle_parent_job_finished(id, parent_child_status, *, get_rabbit):
    _config.crud()["handle_parent_job_finished"](
        id=id, parent_child_status=parent_child_status, get_rabbit=get_rabbit
    )



from copy import deepcopy
from datetime import datetime, timezone
from elody.object_configurations.base_object_configuration import (
    BaseObjectConfiguration,
)
from elody.util import flatten_dict
from uuid import uuid4


class ElodyConfiguration(BaseObjectConfiguration):
    SCHEMA_TYPE = "elody"
    SCHEMA_VERSION = 1

    def crud(self):
        crud = {
            "collection": "entities",
            "collection_history": "history",
            "creator": lambda post_body, **kwargs: self._creator(post_body, **kwargs),
            "post_crud_hook": lambda **kwargs: self._post_crud_hook(**kwargs),
            "pre_crud_hook": lambda **kwargs: self._pre_crud_hook(**kwargs),
            "sorting": lambda key_order_map, **kwargs: self._sorting(
                key_order_map, **kwargs
            ),
        }
        return {**super().crud(), **crud}

    def document_info(self):
        return {"object_lists": {"metadata": "key", "relations": "type"}}

    def logging(self, flat_document, **kwargs):
        return super().logging(flat_document, **kwargs)

    def migration(self):
        return super().migration()

    def serialization(self, from_format, to_format):
        return super().serialization(from_format, to_format)

    def validation(self):
        return super().validation()

    def _creator(
        self,
        post_body,
        *,
        flat_post_body={},
        document_defaults={},
    ):
        flat_post_body = flat_post_body or flatten_dict(
            self.document_info()["object_lists"],
            post_body if isinstance(post_body, dict) else {},
        )
        _id = document_defaults.get("_id", str(uuid4()))

        identifiers = (
            post_body.pop("identifiers", []) if isinstance(post_body, dict) else []
        )
        for property in self.document_info().get("identifier_properties", []):
            if identifier := flat_post_body.get(f"metadata.{property}.value"):
                identifiers.append(identifier)

        template = {
            "_id": _id,
            "identifiers": list(
                set(
                    [
                        _id,
                        *identifiers,
                        *document_defaults.pop("identifiers", []),
                    ]
                )
            ),
            "metadata": [],
            "relations": [],
            "schema": {"type": self.SCHEMA_TYPE, "version": self.SCHEMA_VERSION},
        }

        if isinstance(post_body, dict):
            for key, object_list_key in self.document_info()["object_lists"].items():
                if not key.startswith("lookup.virtual_relations"):
                    post_body[key] = self._merge_object_lists(
                        document_defaults.get(key, []),
                        post_body.get(key, []),
                        object_list_key,
                    )
            document = {**template, **document_defaults, **post_body}
        else:
            document = {**template, **document_defaults}
        return document

    def _document_content_patcher(self, *, document, content, overwrite=False, **_):
        object_lists = self.document_info().get("object_lists", {})
        if overwrite:
            document = content
        else:
            for key, value in content.items():
                if key in object_lists:
                    if key != "relations":
                        for value_element in value:
                            for item_element in document[key]:
                                if (
                                    item_element[object_lists[key]]
                                    == value_element[object_lists[key]]
                                ):
                                    document[key].remove(item_element)
                                    break
                    else:
                        for value_element in value:
                            for item_element in document[key]:
                                if (
                                    item_element[object_lists[key]]
                                    == value_element[object_lists[key]]
                                    and item_element["key"] == value_element["key"]
                                ):
                                    document[key].remove(item_element)
                                    break
                    if not document.get(key):
                        document[key] = []
                    document[key].extend(value)
                else:
                    document[key] = value

        return document

    def _post_crud_hook(self, **kwargs):
        pass

    def _pre_crud_hook(
        self, *, crud, timestamp, document={}, audit_override={}, **kwargs
    ):
        if document:
            document = self.__patch_document_unique_value(document)
            document = self.__patch_document_audit_info(
                crud, document, timestamp, audit_override
            )
            document = self._sanitize_document(
                document=document,
                object_list_name="metadata",
                object_list_value_field_name="value",
            )
            document = self._sanitize_document(
                document=document,
                object_list_name="relations",
                object_list_value_field_name="key",
            )
            document = self._sort_document_keys(document)
        return document

    def _sanitize_document(
        self, *, document, object_list_name, object_list_value_field_name, **kwargs
    ):
        sanitized_document = super()._sanitize_document(document=document, **kwargs)
        for element in document[object_list_name]:
            if not element.get(object_list_value_field_name):
                sanitized_document[object_list_name].remove(element)
        for element in sanitized_document[object_list_name]:
            value = element[object_list_value_field_name]
            if isinstance(value, str):
                lines = value.splitlines()
                value = "\n".join(line.strip() for line in lines).strip()
                element[object_list_value_field_name] = value.strip()
            for key, value in deepcopy(element).items():
                if not value:
                    del element[key]
        return sanitized_document

    def __patch_document_audit_info(self, crud, document, timestamp, audit_override):
        if self._is_request_from_internal_service():
            return document

        document.update({f"date_{crud}d": timestamp})
        if email := self._get_user_context_id():
            label = f"{crud}d_by" if crud == "create" else "last_editor"
            document.update({label: email})
        document.update({key: value for key, value in audit_override.items() if value})
        return document

    def __patch_document_unique_value(self, document):
        if unique_field := self.document_info().get("unique_field"):
            flat_document = flatten_dict(
                self.document_info().get("object_lists", {}), document
            )
            if value := flat_document.get(unique_field):
                document["unique_field"] = f"{document['type']}:{value}"
        return document

    def _sorting(self, key_order_map, **_):
        addFields, sort = {}, {}
        for key, order in key_order_map.items():
            if key == "order":
                addFields.update(
                    {
                        key: {
                            "$arrayElemAt": [
                                {
                                    "$map": {
                                        "input": {
                                            "$filter": {
                                                "input": "$relations",
                                                "as": "relation",
                                                "cond": {
                                                    "$eq": [
                                                        {
                                                            "$arrayElemAt": [
                                                                "$$relation.metadata.key",
                                                                0,
                                                            ]
                                                        },
                                                        key,
                                                    ]
                                                },
                                            }
                                        },
                                        "as": "relation",
                                        "in": {
                                            "$arrayElemAt": [
                                                "$$relation.metadata.value",
                                                0,
                                            ]
                                        },
                                    }
                                },
                                0,
                            ]
                        }
                    }
                )
            elif key not in ["date_created", "date_updated", "last_editor"]:
                addFields.update(
                    {
                        key: {
                            "$arrayElemAt": [
                                {
                                    "$map": {
                                        "input": {
                                            "$filter": {
                                                "input": "$metadata",
                                                "as": "metadata",
                                                "cond": {
                                                    "$eq": ["$$metadata.key", key]
                                                },
                                            }
                                        },
                                        "as": "metadata",
                                        "in": {"$toLower": "$$metadata.value"},
                                    }
                                },
                                0,
                            ]
                        }
                    }
                )
            sort.update({key: order})
        pipeline = []
        if addFields:
            pipeline.append({"$addFields": addFields})
        pipeline.append({"$sort": sort})
        return pipeline



from configuration import get_object_configuration_mapper  # pyright: ignore
from elody.error_codes import ErrorCode, get_error_code, get_read
from elody.util import flatten_dict
from serialization.serialize import serialize  # pyright: ignore
from werkzeug.exceptions import NotFound


def generate_filter_key_and_lookup_from_restricted_key(key):
    if (keys := key.split("@", 1)) and len(keys) == 1:
        return key, {}

    local_field = keys[0]
    document_type, key = keys[1].split("-", 1)
    collection = (
        get_object_configuration_mapper().get(document_type).crud()["collection"]
    )
    lookup = {
        "from": collection,
        "local_field": local_field,
        "foreign_field": "identifiers",
        "as": f"__lookup.virtual_relations.{document_type}",
    }
    return f"{lookup['as']}.{key}", lookup


def get_content(item, request, content):
    return serialize(
        content,
        type=item.get("type"),
        from_format=serialize.get_format(
            (request.view_args or {}).get("spec", "elody"), request.args
        ),
        to_format=get_object_configuration_mapper().get(item["type"]).SCHEMA_TYPE,
    )


def get_flat_item_and_object_lists(item):
    config = get_object_configuration_mapper().get(item["type"])
    object_lists = config.document_info().get("object_lists", {})
    return flatten_dict(object_lists, item), object_lists


def get_item(storage_manager, user_context_bag, view_args) -> dict:
    view_args = view_args or {}
    if id := view_args.get("id"):
        resolve_collections = user_context_bag.get("collection_resolver")
        collection = (
            get_object_configuration_mapper()
            .get(view_args.get("type"))
            .crud()
            .get("collection")
        )
        collections = resolve_collections(
            collection=view_args.get("collection", collection), id=id
        )
        for collection in collections:
            if item := storage_manager.get_db_engine().get_item_from_collection_by_id(
                collection, id
            ):
                return item

    raise NotFound(
        f"{get_error_code(ErrorCode.ITEM_NOT_FOUND, get_read())} | id:{id} - Item with id {id} does not exist."
    )
